/*
 * C module for gathering the DMI data
 */

static char dmi_module_documentation[] =
"Provides access for retrieval of the DMI data on x86 machines\n\n"
"$Id: dmimodule.c,v 1.11 2001/08/12 06:47:00 gafton Exp $"
;

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

/* Python module */
#include "Python.h"

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

struct dmi_header
{
    u_int8_t type;
    u_int8_t length;
    u_int16_t handle;
};

static char *
dmi_string(struct dmi_header *dm, u_int8_t s)
{
    u_int8_t *bp = (u_int8_t *) dm;
    if (!s) return "";
    
    bp += dm->length;
    while (s > 1) {
	bp += strlen(bp);
	bp++;
	s--;
    }
    return bp;
}

static char *
dmi_chassis_type(u_int8_t code)
{
    static char *chassis_type[]={
	"",
	"Other",
	"Unknown",
	"Desktop",
	"Low Profile Desktop",
	"Pizza Box",
	"Mini Tower",
	"Tower",
	"Portable",
	"Laptop",
	"Notebook",
	"Hand Held",
	"Docking Station",
	"All in One",
	"Sub Notebook",
	"Space-saving",
	"Lunch Box",
	"Main Server Chassis",
	"Expansion Chassis",
	"SubChassis",
	"Bus Expansion Chassis",
	"Peripheral Chassis",
	"RAID Chassis",
	"Rack Mount Chassis",
	"Sealed-case PC",
    };
    code &= ~0x80;
	   
    if(code>0x18)
	return "";
    return chassis_type[code];
			 
}

static char *
dmi_processor_type(u_int8_t code)
{
    static char *processor_type[]={
	"",
	"Other",
	"Unknown",
	"Central Processor",
	"Math Processor",
	"DSP Processor",
	"Video Processor"
    };
    
    if(code == 0xFF)
	return "Other";
	       
    if (code > 0xA1)
	return "";
    return processor_type[code];
}

static char *
dmi_processor_family(u_int8_t code)
{
    static char *processor_family[]={
	"",
	"Other",
	"Unknown",
	"8086",
	"80286",
	"Intel386 processor",
	"Intel486 processor",
	"8087",
	"80287",
	"80387",
	"80487",
	"Pentium processor Family",
	"Pentium Pro processor",
	"Pentium II processor",
	"Pentium processor with MMX technology",
	"Celeron processor",
	"Pentium II Xeon processor",
	"Pentium III processor",
	"M1 Family",
	"M1","M1","M1","M1","M1","M1", /* 13h - 18h */
	"K5 Family",
	"K5","K5","K5","K5","K5","K5", /* 1Ah - 1Fh */
	"Power PC Family",
	"Power PC 601",
	"Power PC 603",
	"Power PC 603+",
	"Power PC 604",
    };
    
    if(code == 0xFF)
	return "Other";
	
    if (code > 0x24)
	return "";
    return processor_family[code];
}



/* define the DMI object data structures */
typedef struct {
    PyObject_HEAD
    PyObject 	*info; /* dmi information */
    PyObject	*table; /* dmi table */
} dmi_object;

static PyObject *AccessError;

/* Parse the DMI table */
static int
dmi_table(PyObject *self,
	  int fd, u_int32_t base, int len, int num)
{
    char *buf = malloc(len);
    struct dmi_header *dm;
    u_int8_t *data;
    int i = 0;

    if (lseek(fd, (long) base, 0) == -1) {
	PyErr_SetString(AccessError, "dmi: lseek");
	return 1;
    }
    if (read(fd, buf, len) != len) {
	PyErr_SetString(AccessError, "dmi: read");
	return 1;
    }
    data = buf;
    while (i < num) {
	u_int32_t u;
	PyObject *entry;
	PyObject *tlist;
	const char *entry_name = NULL;
	
	entry = PyDict_New();
	if (entry == NULL) {
	    PyErr_SetString(PyExc_MemoryError, "table item data");
	    return 1;
	}
	
	dm = (struct dmi_header *) data;
	switch (dm->type) {
	    char *s1, *s2;
	    
	    case 0:
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("BIOS Information Block"));
		PyDict_SetItemString(entry, "vendor",
				     PyString_FromString(dmi_string(dm, data[4])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_string(dm, data[5])));
		PyDict_SetItemString(entry, "release",
				     PyString_FromString(dmi_string(dm, data[8])));
		entry_name = "bios";
		break;

	    case 1:
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("System Information Block"));
		PyDict_SetItemString(entry, "vendor",
				     PyString_FromString(dmi_string(dm, data[4])));
		PyDict_SetItemString(entry, "product",
				     PyString_FromString(dmi_string(dm, data[5])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_string(dm, data[6])));
		s1 = strdup(dmi_string(dm, data[7]));
		if ((s2 = strchr(s1, 0xff)))
		    *s2 = 0;	
		PyDict_SetItemString(entry, "serial", PyString_FromString(s1));
		entry_name = "system";
		break;

	    case 2:
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("Board Information Block"));
		PyDict_SetItemString(entry, "vendor",
				     PyString_FromString(dmi_string(dm, data[4])));
		PyDict_SetItemString(entry, "product",
				     PyString_FromString(dmi_string(dm, data[5])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_string(dm, data[6])));
		s1 = strdup(dmi_string(dm, data[7]));
		if ((s2 = strchr(s1, 0xff)))
		    *s2 = 0;	
		PyDict_SetItemString(entry, "serial", PyString_FromString(s1));
		entry_name = "board";
		break;
	    
	    case 3:
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("Chassis Information Block"));
		PyDict_SetItemString(entry, "vendor",
				     PyString_FromString(dmi_string(dm, data[4])));
		PyDict_SetItemString(entry, "product",
				     PyString_FromString(dmi_chassis_type(data[5])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_string(dm, data[6])));
		s1 = strdup(dmi_string(dm, data[7]));
		if ((s2 = strchr(s1, 0xff)))
		    *s2 = 0;	
		PyDict_SetItemString(entry, "serial", PyString_FromString(s1));
		PyDict_SetItemString(entry, "asset",
				     PyString_FromString(dmi_string(dm, data[8])));
		entry_name = "chassis";
		break;

	    case 4:
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("Processor Information Block"));
		PyDict_SetItemString(entry, "socket",
				     PyString_FromString(dmi_string(dm, data[4])));
		PyDict_SetItemString(entry, "type",
				     PyString_FromString(dmi_processor_type(data[5])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_processor_family(data[6])));
		PyDict_SetItemString(entry, "vendor",
				     PyString_FromString(dmi_string(dm, data[7])));
		PyDict_SetItemString(entry, "version",
				     PyString_FromString(dmi_string(dm, data[0x10])));
		entry_name = "processor";
		if (dm->length <= 0x20)
		    break;
		PyDict_SetItemString(entry, "serial",
				     PyString_FromString(dmi_string(dm, data[0x20])));
		PyDict_SetItemString(entry, "asset",
				     PyString_FromString(dmi_string(dm, data[0x21])));
		PyDict_SetItemString(entry, "partno",
				     PyString_FromString(dmi_string(dm, data[0x22])));
		break;

	    case 11:
		tlist = PyList_New(0);	    
		PyDict_SetItemString(entry, "name",
				     PyString_FromString("OEM Data Block"));
		for (u = 0; u < data[4]; u++)
		    PyList_Append(tlist, Py_BuildValue("s", dmi_string(dm, u)));
		PyDict_SetItemString(entry, "data", tlist);
		entry_name = "oem";
		break;
	}

	if (entry_name != NULL && dm->length &&
	    !PyMapping_HasKeyString(self, (char *)entry_name)) {
	    PyDict_SetItemString(entry, "type", Py_BuildValue("i", dm->type));
	    /* And add this new dictionary to the list */
	    PyDict_SetItemString(self, (char *) entry_name, entry);
	} else {
	    Py_DECREF(entry);
	}
	
	data += dm->length;
	while (*data || data[1])
	    data++;      
	data += 2;
	i++;
	entry_name = NULL;
    }
    free(buf);
    return 0;
}


static char key[8] = { 'R', 'S', 'D', ' ', 'P', 'T', 'R', ' ' };

char zot[16];

int
parse_dmi_data(PyObject *self)
{
    unsigned char buf[20];
    PyObject *info;
    
    int fd = open("/dev/mem", O_RDONLY);
    long fp = 0xE0000L;

    if (fd == -1) {
	PyErr_SetString(AccessError, "Could not open /dev/mem for reading");
	return 1;
    }
    if (lseek(fd, fp, 0) == -1) {
	PyErr_SetString(AccessError, "Could not seek on /dev/mem to DMI region");
	return 1;
    }
    
    info = PyDict_New();
    if (info == NULL) {
	PyErr_SetString(PyExc_MemoryError, "memory for info structure");
	return 1;
    }
    PyDict_SetItemString(self, "info", info);
    
    fp -= 16;

    while (fp < 0xFFFFF) {
	fp += 16;
	if (read(fd, buf, 16) != 16)
	    return -1;
	if (memcmp(buf, "_SM_", 4) == 0)
	    PyDict_SetItemString(info, "SMBIOS", Py_BuildValue("i", 1));
	if (memcmp(buf, "_SYSID_", 7) == 0)
	    PyDict_SetItemString(info, "SYSID", Py_BuildValue("i", 1));
	if (memcmp(buf, "_DMI_", 5) == 0) {
	    char buffer[100];
	    int ret;
	    
	    u_int16_t num = buf[13] << 8 | buf[12];
	    u_int16_t len = buf[7] << 8 | buf[6];
	    u_int32_t base = buf[11] << 24 | buf[10] << 16 | buf[9] << 8 | buf[8];

	    snprintf(buffer, sizeof(buffer), "%d.%d", buf[14] >> 4, buf[14] & 0x0F);	    
	    PyDict_SetItemString(info, "DMI", Py_BuildValue("s", buffer));

	    ret = dmi_table(self, fd, base, len, num);
	    if (ret != 0)
		return ret;
	}
	if (memcmp(buf, "$PnP", 4) == 0)
	    PyDict_SetItemString(info, "PNP BIOS", Py_BuildValue("i", 1));
	if (memcmp(buf, key, 3) == 0) {
	    int a;
	    unsigned char sum = 0;

	    if (buf[15] == 0) {
		PyDict_SetItemString(info, "RSD PTR", Py_BuildValue("l", fp));
	    }
	    /* snprintf(FOO, 6, "%s", buf+9) */
	    read(fd, buf + 16, 4);
	    lseek(fd, -4, 1);
	    for (a = 0; a < 20; a++)
		sum += buf[a];
	    if (sum == 0) /* got our OEM */
		PyDict_SetItemString(info, "OEM", Py_BuildValue("s#", buf+9, 6));
	}
    }
    close(fd);

    return 0;			/* avoid fascist compiler warning */
}


/* Create a new DMI object */
static char dmi_object_new_doc[] =
"Creates and initializes a new DMI object";
static PyObject *
dmi_object_new(PyObject *self, PyObject *args)
{
    int ret;
    
    PyObject *new_self;
    
    if (!PyArg_ParseTuple(args, "")) return NULL;

    new_self = PyDict_New();
    if (new_self == NULL)	
	return NULL;

    ret = parse_dmi_data(new_self);
    if (ret != 0) {
	Py_DECREF(new_self);
	if (ret < 0)
	    PyErr_SetString(AccessError, "Could not parse the DMI data");
	return NULL;
    }
       
    return (PyObject *) new_self;
};

static struct PyMethodDef dmi_methods[] = {
    { "DMI", dmi_object_new, METH_VARARGS, dmi_object_new_doc},
    { NULL, NULL },
};

/* Main entry point for the module */
void initdmi()
{
    PyObject *m, *d;
    
    m = Py_InitModule3("dmi", dmi_methods, dmi_module_documentation);
    d = PyModule_GetDict(m);

    AccessError = PyErr_NewException("dmi.AccessError", PyExc_OSError, NULL);
    PyDict_SetItemString(d, "AccessError", AccessError);    
}
