#!/usr/bin/python -u
#
# Part of the up2date upgrader, access to the installed RPM database
#
# veillard@redhat.com
#
"""Implements a class providing access to an installed RPM database.
   Goal is to isolate primitive accesses to keep read-only access to
   the database as much as possible or or to target a group of databases
   on a set of machines."""

import os
import sys
import rpm

import upLog

debug = 0

class rpmInstalledBase:
    """This is the base class for a database exposing a set of installed
       RPMs. It will keep the database opened read-only unless during
       operations requiring write access."""
    def __init__(self):
        #
	# Nothing major here, db is opened only on access
	#
        self.dbpath = "/"
	self.state = 0
	self.db = None
	self.vendor = None
	self.distribution = None
	self.release = None
	self.contentCache = None
	self.pcache = {}
	self.release = None
	self.fd = -1

    #
    # The callback from an RPM runTransaction
    # Let's face it it's seriously ugly, but in a critical path
    #
    def rpmCallback(self, what, amount, total, hdr, path):
        if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
            fileName = "%s/%s-%s-%s.%s.rpm" % (path,
                                               hdr['name'],
                                               hdr['version'],
                                               hdr['release'],
                                               hdr['arch'])
            try:
                self.fd = os.open(fileName, os.O_RDONLY)
            except:
                upLog.error("Error opening %s" % fileName)
            
            return self.fd
        elif what == rpm.RPMCALLBACK_INST_START:
	    upLog.progress("%s-%s-%s.%s.rpm" % (hdr['name'],
                                       hdr['version'],
                                       hdr['release'],
                                       hdr['arch']))
        elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
            os.close(self.fd)

    #
    # Close the rpmdb database.
    #
    def _closeRpmdb(self):
        if self.db == None:
	    return

	upLog.debug("openRpmdb: closing to reopen")

	# self.db.close()
	#
	# Doesn't exist seems the only way is to dereference the
	# object which will close the file descriptors as a side
	# effect ... Neat !
	del(self.db)
	self.db = None
	self.contentCache = None
	self.pcache = {}

    #
    # Open the rpmdb database. the only trick is to get the
    # rpmlib to accept a different path which is done using the
    # ugly addMacro("_dbpath", ...) and popped out once done
    #
    def _openRpmdb(self, dbpath = None, state = 0):
        if self.db != None:
	    if dbpath and dbpath == self.dbpath and self.state == state:
		return 1
	    if dbpath == None and self.state == state:
	        return 1
            self._closeRpmdb()
	    
	if dbpath == None:
	   dbpath = self.dbpath
	if dbpath == None:
	    upLog.error("openRpmdb: metadata base path not defined !")
	    return 0

	if state == 0:
	    upLog.debug("trying to open metadata base %s read-only" % dbpath)
	else:
	    upLog.debug("trying to open metadata base %s read-write" % dbpath)

	if dbpath != '/':
	    rpm.addMacro("_dbpath", dbpath)
	try:
	    self.db = rpm.opendb(state)
	except rpm.error:
	    sys.stderr.write('failed to open metadata base '+ dbpath +'\n')
	    rpm.delMacro("_dbpath")
	    return 0
	if dbpath != '/':
	    rpm.delMacro("_dbpath")
	self.dbpath = dbpath
	self.state = state

	self.contentCache = None
	self.pcache = {}
	return 1

    #
    # initialize the DB by (re)opening it if necessary
    #
    # returns 0 in case of success -1 in case of failure
    #
    def init(self, state = 0):
        ret = self._openRpmdb(None, state)
	if ret == 1:
	    return 0
	return -1

    #
    # Rebuild the content cache
    #
    # Returns 0 if okay -1 otherwise
    #
    def rebuildCache(self):
	self.pcache = {}
        if self.db == None:
	    self._openRpmdb()
	    if self.db == None:
	        return -1

	index = self.db.firstkey()
	if index == None:
	    upLog.error("Local database seems empty !")
	    return -1

	upLog.debug("rebuilding installed cache")

        nbpkg = 0
	self.contentCache = {}
	while index != None:
	    h = self.db[index]
	    if h and h != None:
	        nbpkg = nbpkg + 1
		name = h['name']
		version = h['version']
		release = h['release']
		arch = h['arch']
		epoch = h['epoch']
		if epoch != None:
		    epoch = "%d" % epoch

		if name == 'redhat-release':
		    self.release = h['version']

		if self.contentCache.has_key(name):
		    dists = self.contentCache[name]
		else:
		    dists = [];
		self.contentCache[name] = dists
		dists.append((name, version, release, epoch, arch))

	    index = self.db.nextkey(index)

        upLog.debug("local base cache rebuilt %d packages" % (nbpkg))

        return 0


    #
    # Search the local cache to find the packages providing a given
    # resource.
    #
    # Returns a list of NVRE
    #
    def getList(self, name):
	if name == None:
	    return []

	if self.contentCache == None:
	    if self.rebuildCache() == -1:
	        return []

	if self.contentCache.has_key(name):
	    list = self.contentCache[name]
	    return list
	
	return []


    #
    # Check if a package is already installed
    # resource.
    #
    # Returns 1 if true, 0 if false and -1 in case of error
    #
    def isInstalled(self, pkg):
        if pkg == None:
	    return -1

        if self.db == None:
	    self._openRpmdb()
	    if self.db == None:
	        return -1

	list = self.getList(pkg[0])

        for index in list:
	    v = index[1]
	    r = index[2]
	    if pkg[1] == v and pkg[2] == r:
		return 1

	return 0

    #
    # Search the rpmdb datbase to find what provides a given resource
    #
    # Returns a list of package descriptions (possibly empty)
    #
    def whatProvides(self, resource):
	if resource == None:
	    return []
        if self.db == None:
	    if self._openRpmdb() == 0:
	        return []
        if self.db == None:
	    return []

        if self.pcache.has_key(resource):
            return self.pcache[resource]
	
	if resource[0] == '/':
	    try:
		indexes = self.db.findbyfile(resource)
	    except:
	        indexes = []
	else:
	    try:
		indexes = self.db.findbyname(resource)
	    except:
	        indexes = []
	    if len(indexes) == 0:
		try:
		    indexes = self.db.findbyprovides(resource)
		except:
		    indexes = []

	if indexes == None or len(indexes) == 0:
	    return []

        list = []
	for index in indexes:
	    h = self.db[index]
	    n = h['name']
	    v = h['version']
	    r = h['release']
	    e = h['epoch']
	    if e == None:
	        e = ''
	    a = h['arch']
	    list.append((n, v, r, e, a, ''))
	
	self.pcache[resource] = list
	return list
	
    #
    # Check a priori if a package, defined by its NVRE is an
    # update of an installed package.
    #
    # Returns 1 if yes, 0 if older, -1 if not installed, -2 in case of error
    #
    def isUpdate(self, name, version, release, epoch = None):
        global debug

        found_older = 0
        found_equal = 0
	found_newer = 0

	if name == None or version == None or release == None:
	    return -2

	#
	# labelCompare expects None for an absence of epoch
	#
	if epoch == "":
	    epoch = None

        if self.db == None:
	    if self._openRpmdb() == 0:
	        return -2
        if self.db == None:
	    return -2
        
	list = self.getList(name)

        for index in list:
	    oldv = index[1]
	    oldr = index[2]
	    olde = index[3]

	    upLog.debug("found %s-%s-%s:%s" % (name, oldv, oldr, olde), 1)

	    cmp = rpm.labelCompare((olde, oldv, oldr), \
	                           (epoch, version, release))
	    upLog.debug("  result: %s" % (cmp), 1)

	    if cmp < 0:
	        found_older = found_older + 1
	    elif cmp > 0:
	        found_newer = found_newer + 1
	    elif cmp == 0:
	        found_equal = found_equal + 1

	upLog.debug("total: %s older, %s equal, %s newer" % \
	          (found_older, found_equal, found_newer), 1)

        #
	# @@ whether to upgrade or not is unclear, this will have
	# to be hooked to the policy stuff.
	#
	if found_older + found_equal + found_newer > 1:
	    upLog.warning("Warning found %d %s packages" % 
	                  (found_older + found_equal + found_newer, name))

	if found_equal > 0:
	    return 0
	if found_newer > 0:
	    return 0
	if found_older > 0:
	    return 1
        return -1

    #
    # Check if a package, defined by is header is an update of an
    # installed package.
    #
    # Returns 1 if yes, 0 if no, -1 in case of error
    #
    def checkUpdate(self, header, force):
        if header == None:
	    return -1

        #
	# Hum , though not perfect the best seems to extract the
	# NVRE from the header and call isUpdate() . Running a single
	# header transaction is likely to be far more costly and harder
	# to interpret. But checkUpdate() should be differenciated in
	# case this get provided at the python API level.
	#
	try:
	    name = header['name']
	    version = header['version']
	    release = header['release']
	    epoch = header['epoch']
	except:
	    return -1

        return self.isUpdate(name, version, release, epoch)

    #
    # Check a transaction set made of a list of RPM headers
    #
    # Returns the dependancies which need to be solved
    #
    def checkTransaction(self, headers):
        if headers == None:
	    return None

        if self.db == None:
	    if self._openRpmdb() == 0:
	        return None
        if self.db == None:
	    return None

	ts = rpm.TransactionSet(self.dbpath, self.db)
        if ts == None:
	    return None

        nb = 0
        for h in headers:
	    # TODO distinguish between upgrade and installs here
	    upLog.debug("Adding %s-%s-%s to transaction" % \
		      (h['name'], h['version'], h['release']), 1)
	    ts.add(h, h, "u")
	    nb = nb + 1

        upLog.debug("Checking transaction on local base: %s packages" % nb)

	ts.order()
	deps = ts.depcheck()
	del(ts)

	if deps:
	    upLog.debug("RPM check returned %d deps." % len(deps))
	    for err in deps:
	        (byName, byVersion, byRelease) = err[0]
		(needsName, needsVersion) = err[1]
		flags = err[2]
		str = "   %s-%s-%s requires %s " % (byName, byVersion,
		      byRelease, needsName)
		flags = flags & 0xFF
		if flags != 0 and needsVersion != None:
		    if flags & rpm.RPMSENSE_LESS:
			str = str + "<"
		    if flags & rpm.RPMSENSE_GREATER:
			str = str + ">"
		    if flags & rpm.RPMSENSE_EQUAL:
			str = str + "="
		    if flags & rpm.RPMSENSE_SERIAL:
			str = str + "S"
		    str = str + needsVersion
		upLog.debug(" %s" % str)

	
	return deps

    #
    # Run a transaction set: this implies opening the localbase
    # for write access, running the transaction, switching back to
    # read-only, cleaning up the cache and returning the error set if any
    #
    # Returns the error set if any
    #
    def runTransaction(self, headers, directory, flags = 0, callback = None):
        if headers == None:
	    return None

        if self._openRpmdb(None, 1) != 1:
	    upLog.error("Failed re-opening the local base in write mode")
	    return None
	    
	if flags == None:
	    flags = 0
	if callback == None:
	    callback = self.rpmCallback

	ts = rpm.TransactionSet(self.dbpath, self.db)
        if ts == None:
	    return None

        upLog.debug("----------------")
	upLog.debug("Running transaction on local base")
        for h in headers:
	    # TODO distinguish between upgrade and installs here
	    ts.add(h, h, "u")
	    upLog.debug("    upgrade %s-%s-%s" %
		      (h['name'], h['version'], h['release']))
        upLog.debug("----------------")
	
	#
	# TODO: the RPMPROB_FILTER_REPLACEOLDFILES should be a policy setting
	#
	ts.order()
	ret = ts.run(flags, rpm.RPMPROB_FILTER_REPLACEOLDFILES,
	             callback, directory)

	self.contentCache= None
	self.pcache = {}

	#
	# Close the database ... it will be reopened on access and it
	# and be required to have a closed DB for some of the post-install
	# scripts
	#
        self._closeRpmdb()

	return ret

    #
    # Guess the distribution and report statistics
    #
    def guessRpmDistribution(self):
        frequent = {}

        if self.db == None:
	    if self._openRpmdb() == 0:
	        return ('none', 'none')

        if self.db == None:
	    return ('none', 'none')

	index = self.db.firstkey()
	if index == None:
	    return ('none', 'none')

        nbpkg = 0;
	release = None
	while index != None:
	    h = self.db[index]
	    if h and h != None:
	        nbpkg = nbpkg + 1
		dist = h['distribution']
		vend = h['vendor']

		if h['name'] == 'redhat-release':
		    release = h['version']

		if frequent.has_key(vend):
		    dists = frequent[vend]
		else:
		    dists = {};
		frequent[vend] = dists

		if dists.has_key(dist):
		    occur = dists[dist]
		else:
		    occur = 0
		dists[dist] = occur + 1

	    index = self.db.nextkey(index)

	count = 0;
	ret = ('none', 'none')
	for vend in frequent.keys():
	    vends = frequent[vend]
	    for dist in vends.keys():
	        if vends[dist] > count:
		    ret = (vend, dist)
		    count = vends[dist]
		    
        upLog.progress("Default distribution : %s(%s)" % ret)
	upLog.progress("\towning %d of %d installed packages" % (count, nbpkg))
	if release == None:
	    upLog.progress("\tVersion of Red Hat not found")
	else:
	    upLog.progress("\tVersion of Red Hat: %s" % release)

	self.vendor = ret[0]
	self.distribution = ret[1]
	self.release = release
	return ret

#
# Small test routine to do --whatprovides
# a given symbol using the installed database
#
def main():
    global debug
    localDatabase = None

    args = sys.argv[1:]
    while args:
        if args[0] == '--whatprovides' or args[0] == '-whatprovides':
	    args = args[1:]
	    if args[0] == None:
		return

	    upLog.debug("Searching resource", args[0])
	    if localDatabase == None:
		localDatabase = rpmInstalledBase()
	    if localDatabase == None:
		sys.exit(-1)

	    list = localDatabase.whatProvides(args[0])
	    print "%d provides found" % len(list)
	    for h in list:
		print "%s-%s-%s" % (h[rpm.RPMTAG_NAME],
				    h[rpm.RPMTAG_VERSION],
				    h[rpm.RPMTAG_RELEASE])

        if args[0] == '--list' or args[0] == '-list':
	    args = args[1:]
	    if args[0] == None:
		return

	    upLog.debug("Searching resource", args[0])
	    if localDatabase == None:
		localDatabase = rpmInstalledBase()
	    if localDatabase == None:
		sys.exit(-1)

	    list = localDatabase.getList(args[0])
	    print "%d packages found" % len(list)
	    for h in list:
		print "%s-%s-%s" % (h[0], h[1], h[2])

        elif args[0] == '--isinstalled' or args[0] == '-isinstalled':
	    args = args[1:]
	    name = args[0]
	    args = args[1:]
	    version = args[0]
	    args = args[1:]
	    release = args[0]

	    if localDatabase == None:
		localDatabase = rpmInstalledBase()
	    if localDatabase == None:
		sys.exit(-1)

            ret = localDatabase.isInstalled((name, version, release, "", None))
	    print "result:", ret


        elif args[0] == '--isupdate' or args[0] == '-isupdate':
	    args = args[1:]
	    name = args[0]
	    args = args[1:]
	    version = args[0]
	    args = args[1:]
	    release = args[0]

	    if localDatabase == None:
		localDatabase = rpmInstalledBase()
	    if localDatabase == None:
		sys.exit(-1)

            ret = localDatabase.isUpdate(name, version, release)
	    print "result:", ret

        elif args[0] == '--verbose' or args[0] == '-v':
	    debug = debug + 1

        elif args[0] == '--guess' or args[0] == '-guess':
	    upLog.debug("Guessing distrib")
	    if localDatabase == None:
		localDatabase = rpmInstalledBase()
	    if localDatabase == None:
		sys.exit(-1)

	    localDatabase.guessRpmDistribution()

        else:
            sys.stderr.write('unknown argument '+ args[0] + '\n')
	    sys.stderr.write('use --guess or --whatprovides res\n')
	    sys.exit(-1)
        args = args[1:]


if __name__ == "__main__":
    main()
