#!/usr/bin/python -u
#
# Part of the up2date upgrader, metadata lookup
#
# veillard@redhat.com
#
"""Implements an RPM metadata resolver class. Assuming a resource
   is missing it will try to find the package providing it using
   the rpmdb metadata database if installed."""

import os
import sys
import string
import rpm
import posixpath

import upLog

debug = 0

class rpmDbMetadata:
    """This is the base class for a metadata resolver. It currently
       tries to open the rpmdb metadata database if installed, and 
       lookup informations from it."""
    def __init__(self, dbpath):
        self.dbpath = dbpath
	self.db = None
	self.pcache = {}
	self.ocache = None

    #
    # Open the rpmdb extra database. the only trick is to get the
    # rpmlib to accept a different path which is done using the
    # ugly addMacro("_dbpath", ...) and popped out once done
    #
    def _openRpmdb(self, dbpath = None):
        if self.db:
	    if dbpath and dbpath == self.dbpath:
		upLog.error("openRpmdb: metadata base already opened !")
		return 1
	if dbpath == None:
	   dbpath = self.dbpath
	if dbpath == None:
	    upLog.error("openRpmdb: metadata base path not defined !")
	    return 0
        upLog.debug("trying to open metadata base %s" % dbpath)
	
	#
	# try to avoid RPM messages
	#
	if not posixpath.isdir(dbpath):
	    upLog.debug("No such directory: %s" % dbpath)
	    return 0

        rpm.addMacro("_dbpath", dbpath)
	try:
	    self.db = rpm.opendb(0)
	except rpm.error:
	    sys.stderr.write("failed to open metadata base %s" % dbpath)
	    rpm.delMacro("_dbpath")
	    return 0
	rpm.delMacro("_dbpath")
	self.dbpath = dbpath
	self.pcache = {}
	self.ocache = None
	self.olist = None
	return 1

    #
    # Initialize the metadata database
    #
    # Returns 0 in case of success, -1 in case of error
    #
    def init(self, path = None):
        ret = self._openRpmdb(path)
	if ret == 1:
	   return 0
	return -1
        
    #
    # Invalidate the rpmdb extra database.
    #
    def invalidate(self):
        if self.db != None:
	    del(self.db)
	    self.db = None
	self.pcache = {}
	self.ocache = None

    #
    # Rebuild the obsoletes cache
    #
    # Returns 0 if okay -1 otherwise
    #
    def rebuildCache(self):
        if self.db == None:
	    self._openRpmdb()
	    if self.db == None:
	        return -1

	index = self.db.firstkey()
	if index == None:
	    upLog.error("Local database seems empty !")
	    return -1

	upLog.debug("rebuilding obsolete cache")

        nbpkg = 0
	self.ocache = {}
	self.olist = []
	while index != None:
	    h = self.db[index]
	    if h and h != None:
		obsoletes = h['obsoletes']
		if obsoletes != None:
		    for o in obsoletes:
		        self.olist.append(o)
			if self.ocache.has_key(o):
			    list = self.ocache[o]
			else:
			    list = [];
			    nbpkg = nbpkg + 1
			self.ocache[o] = list
			n = h['name']
			v = h['version']
			r = h['release']
			e = h['epoch']
			if e == None:
			    e = ''
			a = h['arch']
			list.append((n, v, r, e, a, ''))

	    index = self.db.nextkey(index)

        upLog.debug("local base obsolete cache rebuilt %d obsoletes" % (nbpkg))

        return 0

    #
    # Search the rpmdb datbase to find what provides a given resource
    #
    # Returns a list of package descriptions (possibly empty)
    #
    def whatProvides(self, resource):
	if resource == None:
	    return []
        if self.db == None:
	    if self._openRpmdb() == 0:
	        return []
        if self.db == None:
	    return []

        if self.pcache.has_key(resource):
            return self.pcache[resource]
	
	if resource[0] == '/':
	    indexes = self.db.findbyfile(resource)
	else:
	    indexes = self.db.findbyprovides(resource)
	if indexes == None or len(indexes) == 0:
	    return []

        list = []
	for index in indexes:
	    h = self.db[index]
	    n = h['name']
	    v = h['version']
	    r = h['release']
	    e = h['epoch']
	    if e == None:
	        e = ''
	    a = h['arch']
	    list.append((n, v, r, e, a, ''))
	
	self.pcache[resource] = list
	return list
	
    #
    # Search the rpmdb database to find packages which obsoletes a
    # given resource
    #
    # Returns a list of package descriptions (possibly empty)
    #
    def whatObsoletes(self, resource):
	if resource == None:
	    return []

        if self.db == None:
	    if self._openRpmdb() == 0:
	        return []

        if self.db == None:
	    return []

	if self.ocache == None:
	    if self.rebuildCache() == -1:
	        return []

	if self.ocache.has_key(resource):
	    list = self.ocache[resource]
	    return list
        return []

    #
    # Search the rpmdb database to find what resources are obsoleted
    # by a given package
    #
    # Returns a list of obsoleted resource names
    #
    def obsoletesList(self, name):
        list = []
	if name == None:
	    return list

        if self.db == None:
	    if self._openRpmdb() == 0:
	        return list

        if self.db == None:
	    return list

	indexes = self.db.findbyprovides(name)
	for index in indexes:
	    h = self.db[index]
	    obsoletes = h['obsoletes']
	    if obsoletes:
		for x in obsoletes:
		    list.append(x)

        return list

    #
    # Search the rpmdb database to find all resources that are obsoleted
    #
    # Returns a list of obsoleted resource names
    #
    def obsoletesAll(self):
        if self.db == None:
	    if self._openRpmdb() == 0:
	        return []

        if self.db == None:
	    return []

	if self.olist == None:
	    if self.rebuildCache() == -1:
	        return []

	return self.olist

class rpmMetadataSources:
    """This is front-end to a set of metadata sources, it distinguish
       two classes of sources, the local rpmdb and remotes ones, local
       ones are queried first"""
    def __init__(self):
        self.rpmdbSources = []
        self.extraSources = []
	self.initialized = 0


    #
    # Init: will initialize the rpmdb lookups
    #
    # return the number of base opened or -1 in case of error
    #
    def init(self):
        if self.initialized != 0:
	    return -1

	#
	# find all the rpmdb databases
	#
	directories = [ "/usr/lib/rpmdb" ]
	databases = []

	while len(directories) > 0:
	    dir = directories[0]
	    directories = directories[1:]
	    try:
		filenames = os.listdir(dir)
	    except:
	        filenames = []

	    for filename in filenames:
		if filename == "Packages":
		    databases.append(dir)
		elif filename[0] != '.' :
		    path = dir + "/" + filename
		    if posixpath.isdir(path):
			directories.append(path)

        #
	# Add them to the pool
	#
	for db in databases:
	    rpmdb = rpmDbMetadata(db)
	    if rpmdb != None:
	        self.rpmdbSources.append(rpmdb)
	
	self.initialized = 1

        upLog.debug("initialized %s local metadata sources" %
	            len(self.rpmdbSources))

	return len(self.rpmdbSources)

    #
    # Invalidate the rpmdb metadata source(s)
    #
    def invalidate(self):
	for src in self.rpmdbSources:
	    src.invalidate()
	self.rpmdbSources = []
	self.initialized = 0

        
    #
    # Add a generic metadata source, examples coming to mind are
    # direct requests to RHN or rpmfind like services
    #
    # Returns 0 in case of success, -1 in case of error
    #
    def addSource(self, src):
        return -1

    #
    # Add a generic metadata source, like addSource, but put it in front
    # so it's queried first
    #
    # Returns 0 in case of success, -1 in case of error
    #
    def addSourceHead(self, src):
        return -1

    #
    # Search the metadata sources to find what provides a given resource
    #
    # Returns a list of package descriptions (possibly empty)
    #
    def whatProvides(self, resource):
        if self.initialized == 0:
	    self.init()

        ret = []
	for src in self.rpmdbSources:
	    res = src.whatProvides(resource)
	    ret = ret + res

	if len(ret) > 0:
	    return ret

	for src in self.extraSources:
	    res = src.whatProvides(resource)
	    ret = ret + res

        return ret

    #
    # Search the rpmdb database to find packages which obsoletes a
    # given resource
    #
    # Returns a list of package descriptions (possibly empty)
    #
    def whatObsoletes(self, resource):
        if self.initialized == 0:
	    self.init()

        ret = []
	for src in self.rpmdbSources:
	    res = src.whatObsoletes(resource)
	    ret = ret + res

	if len(ret) > 0:
	    return ret

	for src in self.extraSources:
	    res = src.whatObsoletes(resource)
	    ret = ret + res

        return ret

    #
    # Search the rpmdb database to find what resources are obsoleted
    # by a given package
    #
    # Returns a list of obsoleted resource names
    #
    def obsoletesList(self, name):
        if self.initialized == 0:
	    self.init()

	for src in self.rpmdbSources:
	    res = src.obsoletesList(name)
	    if len(res) > 0:
		return res

	for src in self.extraSources:
	    res = src.obsoletesList(name)
	    if len(res) > 0:
		return res

        return []

    #
    # Search the rpmdb database to find all resources that are obsoleted
    #
    # Returns a list of obsoleted resource names
    #
    def obsoletesAll(self):
        if self.initialized == 0:
	    self.init()

        ret = []
	for src in self.rpmdbSources:
	    res = src.obsoletesAll()
	    ret = ret + res

	if len(ret) > 0:
	    return ret

	for src in self.extraSources:
	    res = src.obsoletesAll()
	    ret = ret + res

        return ret


if __name__ == "__main__":
    localMetadata = None

    args = sys.argv[1:]
    while args:
        if args[0] == '--whatprovides' or args[0] == '-whatprovides':
	    args = args[1:]
	    resource = args[0]
	    if debug:
		print "Searching resource", resource

	    if resource == None:
		continue

	    if localMetadata == None:
		localMetadata = rpmMetadataSources()

	    list = localMetadata.whatProvides(resource)
	    print "%d provides found" % len(list)
	    for pkg in list:
		print "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])

	elif args[0] == '--whatobsoletes' or args[0] == '-whatobsoletes':
	    args = args[1:]
	    resource = args[0]
	    if debug:
		print "Searching what obsoletes resource", resource

	    if resource == None:
		continue

	    if localMetadata == None:
		localMetadata = rpmMetadataSources()

	    list = localMetadata.whatObsoletes(resource)
	    print "%d obsoletes found" % len(list)
	    for pkg in list:
		print "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])

	elif args[0] == '--obsoletelist' or args[0] == '-obsoletelist':
	    args = args[1:]
	    resource = args[0]
	    if debug:
		print "Searching what resources are obsoleted by", resource

	    if resource == None:
		continue

	    if localMetadata == None:
		localMetadata = rpmMetadataSources()

	    list = localMetadata.obsoletesList(resource)
	    print "%d obsoletes found" % len(list)
	    for h in list:
		print h

        elif args[0] == '--verbose' or args[0] == '-v':
	    debug = debug + 1

        else:
            sys.stderr.write('unknown argument '+ args[0] + '\n')
	    sys.stderr.write('use --whatprovides res\n')
	    sys.stderr.write('or  --whatobsoletes res\n')
	    sys.stderr.write('or  --obsoletelist res\n')
	    sys.exit(-1)
        args = args[1:]
