#!/usr/bin/python -u
#
# Part of the up2date upgrader, class handling the upgrade package history
#
# veillard@redhat.com
#
"""Implements a class used for bookeeping the upgradeHistory,
   this is used both within the session to find the name and state
   of possible packages, but also in case of needed restart to
   lookup possibly older packages needed anyway for an upgrade."""

import sys
import string

import upLog

debug = 0

UPGRADE_STATUS_UNKNOWN = 0
UPGRADE_STATUS_DONE = 1
UPGRADE_STATUS_TODO = 2
UPGRADE_STATUS_OLD = 3
UPGRADE_STATUS_SKIPPED = 4
UPGRADE_STATUS_ARCH = 5
UPGRADE_STATUS_BROKEN = 6

class upgradeHistory:
    """This is the base class used for bookeeping the upgradeHistory,
       this is used both within the session to find the name and state
       of possible packages, but also in case of needed restart to
       lookup possibly older packages needed anyway for an upgrade."""
    def __init__(self, filename = None):
        #
	# Nothing major here, should allow to init from a file
	#
        self.history = []
	self.filename = filename
	if filename != None:
	    self.loadFromFile(filename)

	upLog.debug("upgradeHistory initialized")

    #
    # Add an entry in the history, won't check for doublons
    #
    # Returns the new entry
    #
    def add(self, pkg, status = UPGRADE_STATUS_UNKNOWN):
        if pkg == None:
	    return None

	upLog.debug("adding %s-%s-%s.%s : %s" %
	          (pkg[0], pkg[1], pkg[2], pkg[4], status))

        ret = [pkg, status]
	self.history.append(ret)
	return ret

    #
    # Update an entry in the history
    #
    # Returns the new entry or the updated one
    #
    def update(self, pkg, status):
        if pkg == None:
	    return None

        ret = None
	for e in self.history:
	    x = e[0]
	    if x is pkg:
	       ret = e
	       break
	    if x[0] == pkg[0] and x[1] == pkg[1] and \
	       x[2] == pkg[2] and x[4] == pkg[4]:
	       ret = e
	       break

        if ret == None:
	    upLog.debug("adding %s-%s-%s.%s : %s" % 
		      (pkg[0], pkg[1], pkg[2], pkg[4], status))
	    ret = [pkg, status]
	    self.history.append(ret)
	else:
	    upLog.debug("updating %s-%s-%s.%s : %s" %
		      (pkg[0], pkg[1], pkg[2], pkg[4], status))
	    ret[1] = status

	return ret

    #
    # search an entry in the history
    #
    # Returns the list of entries corresponding to the criteria
    #
    def search(self, name, version = None, release = None, epoch = None,
               arch = None):
        if name == None:
	    return []

        ret = []
	for e in self.history:
	    x = e[0]
	    if x[0] == name:
	        if version != None and version != x[1]:
		    continue
	        if release != None and release != x[2]:
		    continue
	        if epoch != None and epoch != x[3]:
		    continue
	        if arch != None and arch != x[4]:
		    continue
		ret.append(e)

	return ret
    #
    # Sort an history with an user provided function
    # (will come from the policy)
    #
    def sort(self, f):
        if self.history == None:
	    return

        nlist = map(lambda x, f=f: (f(x[0]), x), self.history)
	nlist.sort()
	self.history = map(lambda (key, x): x, nlist)

    #
    # Lenght of the history
    #
    def len(self):
        if self.history == None:
	    return 0
	return len(self.history)

    #
    # Item in the history
    #
    def item(self, index):
        if self.history == None:
	    return None
	return self.history[index]

    #
    # Need a function to save an history
    #
    def saveToFile(self, filename = None):
        print "Not implemented"

    #
    # Need a function to load an history
    #
    def loadFromFile(self, filename):
        print "Not implemented"


#
# Small test routine to do basic tests
#
def main():
    global debug
    history = None

    args = sys.argv[1:]
    while args:
        if args[0] == '--check' or args[0] == '-check':
	    args = args[1:]
	    if args[0] == None:
		return

	    upLog.debug("Checking resource %s " % args[0])
	    if history == None:
		history = upgradeHistory()
	    if history == None:
		sys.exit(-1)

        elif args[0] == '--verbose' or args[0] == '-v':
	    debug = debug + 1

        else:
            sys.stderr.write('unknown argument '+ args[0] + '\n')
	    sys.stderr.write('use --check name version realease or -v\n')
	    sys.exit(-1)
        args = args[1:]


if __name__ == "__main__":
    main()
