#!/usr/bin/python -u
#
# Part of the up2date upgrader, class defining policies for upgrade
#
# veillard@redhat.com
#
"""Implements the classes defining policies to use when running an RPM
   upgrader. This should include things like packages to not upgrade,
   packages to upgrade in priority, or packages where upgrading by 
   reinstalling a new copy is accepted. Actions can be associated to
   packages upgrade and are defined as scripts embedded in the policy."""

import sys
import string
import re

import upLog

debug = 0

class defaultPolicy:
    """This is the default class defining checking policies 
       for the upgrader. This defines packages to not upgrade,
       packages to upgrade in priority, or packages where upgrading by 
       reinstalling a new copy is accepted. Actions can be associated to
       packages upgrade and are defined as scripts embedded in the policy."""

    def __init__(self):
        #
	# Nothing major here, should allow to init from a file
	#
        self.noupgrade = []
	self.noupgrade.append(re.compile('kernel.*'))
	self.installover = []
        self.force = []
	self.force.append(re.compile('rpmdb-redhat.*'))
        self.priority = []

	#
	# Default scripts: run commands after rpm and lilo upgrades
	#
        self.postcommand = []
	self.postcommand.append((re.compile('^rpm$'), 
	     "import os ; os.system('rpm --rebuilddb')"))
	self.postcommand.append((re.compile('^lilo$'), 
	     "import os ; os.system('lilo')"))

	#
	#this should probably be sorted by regexp lenght at least
	#
	self.priority.append((re.compile('rpmdb-redhat.*'),1000))
	self.priority.append((re.compile('.*compat.*'),500))
	self.priority.append((re.compile('glibc.*'),250))
	self.priority.append((re.compile('rpm.*'),100))
	self.priority.append((re.compile('libc.*'),40))
	self.priority.append((re.compile('ncurses.*'),10))
	self.priority.append((re.compile('readline.*'),10))
	self.priority.append((re.compile('chkconfig.*'),5))
	upLog.debug("defaultPolicy initialized")

    def noupgradeCheck(self, pkg):
        name = pkg[0]
        for x in self.noupgrade:
	    if x.match(name):
		upLog.debug("Found no upgrade rule for %s" % name, 1)
	        return 1
	return 0

    def forceCheck(self, pkg):
        name = pkg[0]
        for x in self.force:
	    if x.match(name):
		upLog.debug("Found force rule for %s" % name, 1)
	        return 1
	return 0

    def installoverCheck(self, pkg):
        name = pkg[0]
        for x in self.installover:
	    if x.match(name):
		upLog.debug("Found install over rule for %s" % name, 1)
	        return 1
	return 0

    def priorityCheck(self, pkg):
        name = pkg[0]
        for x in self.priority:
	    if x[0].match(name):
		upLog.debug("Found priority rule for %s" % name, 1)
	        return x[1]
	return 0

    def postScript(self, pkg):
        name = pkg[0]
        for x in self.postcommand:
	    if x[0].match(name):
		upLog.debug("Found post command rule for %s" % name, 1)
	        return x[1]
	return None

    #
    # Add a rule forbidding the upgrade for a regexp
    #
    def forbidUpgrade(self, regexp):
        if regexp == None:
	    return
	self.noupgrade.append(re.compile(regexp))

    #
    # Add a rule forcing the upgrade for a regexp
    #
    def forceUpgrade(self, regexp):
        if regexp == None:
	    return
	self.noupgrade.append(re.compile(regexp))

    #
    # Add a rule allowing multiple installation of packages
    #
    def allowMultiple(self, regexp):
        if regexp == None:
	    return
	self.installover.append(re.compile(regexp))

    #
    # Add a priority rule
    #
    def setPriority(self, regexp, priority):
        if regexp == None or priority == None:
	    return
	self.priority.insert(0, (re.compile(regexp), priority))


class upgradePolicies:
    """This class defines a registration for policies to apply to
       an upgrade processus. This keep an ordered list of policies
       instances"""

    def __init__(self):
        self.policiesSources = []
	self.initialized = 0
        self.default = defaultPolicy()
	if self.default == None:
	    upLog.error("Error creating the default policies")
	else:
	    self.policiesSources.append(self.default)

    #
    # Register a new policy
    #
    def registerPolicy(self, custom):
        if custom == None:
	    return

	self.policiesSources.append(custom)

    #
    # Disable the default policy
    #
    def disableDefaults(self):
        if self.default == None:
	    return

	self.policiesSources.delete(self.default)
	self.default = None

    #
    # Check a given pkg for rules forbidding its upgrade
    #
    # Returns 1 if the package must not be upgraded
    #
    def noupgradeCheck(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

        for src in self.policiesSources:
	    ret = src.noupgradeCheck(pkg)
	    if ret == 1:
	        return 1
	return 0

    #
    # Check a given pkg for rules forcing its upgrade or installation
    #
    # Returns 1 if the package must be upgraded
    #
    def forceCheck(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

        for src in self.policiesSources:
	    ret = src.forceCheck(pkg)
	    if ret == 1:
	        return 1
	return 0

    #
    # Check a given pkg for rules asking to not upgrade it but install
    # a new instance if a newer is found
    #
    # Returns 1 if the package must not be upgraded but reinstalled
    #
    def installoverCheck(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

        for src in self.policiesSources:
	    ret = src.installoverCheck(pkg)
	    if ret == 1:
	        return 1
	return 0

    #
    # Check the priority for a given pkg. The default priority is
    # 0 but a higher priority can be given to force the package to
    # be installed earlier in the upgrade process.
    #
    # Returns the highest priority for the package as found on the
    #         various sources
    #
    def priorityCheck(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

        prio = 0
        for src in self.policiesSources:
	    ret = src.priorityCheck(pkg)
	    if ret > prio:
	        prio = ret
	return prio

    #
    # Check if a given package is allowed to be upgraded and if yes
    # the associated priority. Combine information from the registered
    # policies sources
    #
    # Returns 0 if not allowed -1 for special conditions, and otherwise
    #      the package priority
    #
    def checkUpgrade(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

	upLog.debug("allowUpgrade: checking %s" % pkg[0], 1)
	if self.noupgradeCheck(pkg):
	    return 0
	if self.installoverCheck(pkg):
	    return -1
	
	ret = self.priorityCheck(pkg)
	if ret != 0:
	    return ret
	return 1

    #
    # Check if a given package has post install scripts defined
    #
    # Returns the list of scripts defined by the policies
    #
    def postScript(self, pkg):
        if pkg == None or pkg[0] == None or pkg[0] == "":
	    return -1

        list = []
        for src in self.policiesSources:
	    ret = src.postScript(pkg)
	    if ret != None:
	        list.append(ret)
	return list

if __name__ == "__main__":
    global debug
    policy = None

    args = sys.argv[1:]
    while args:
        if args[0] == '--check' or args[0] == '-check':
	    args = args[1:]
	    if args[0] == None:
		continue

	    upLog.debug("Checking resource %s" % args[0])
	    if policy == None:
		policy = upgradePolicies()
	    if policy == None:
		sys.exit(-1)

            policy.setPriority("rpm-python", 500)

	    ret = policy.checkUpgrade(args[0])
	    if ret < 0:
	        print "special case"
	    elif ret == 0:
	        print "unauthorized upgrade"
	    elif ret > 1:
	        print "priority package", ret
	    else:
	        print "okay"

        elif args[0] == '--post' or args[0] == '-post':
	    args = args[1:]
	    if args[0] == None:
		continue

	    upLog.debug("Checking post command %s" % args[0])
	    if policy == None:
		policy = upgradePolicies()
	    if policy == None:
		sys.exit(-1)

	    ret = policy.postCmd(args[0])
	    if ret == None:
	        print "No post install command for", args[0]
            else:
	        print "Post install command for %s is '%s'" % (args[0], ret)

        elif args[0] == '--verbose' or args[0] == '-v':
	    debug = debug + 1

        else:
            sys.stderr.write('unknown argument '+ args[0] + '\n')
	    sys.stderr.write('use --check name or -v\n')
	    sys.exit(-1)
        args = args[1:]
