#!/usr/bin/python -u
#
# Part of the up2date upgrader, the core class implementing the
# resolver part of the upgrade system
#
# veillard@redhat.com
#
"""Implements the class in charge of actually running the upgrade
   algorithm based on a set of available informations coming from
   the installed base, the metadata sources, the RHN interface,
   the policies and an history"""

import os
import sys
import string
import rpm
sys.path.append("..")
sys.path.append("../../modules/xmlrpc")
import installedBase
import rhnInterface
import rpmMetadata
import upgradePolicy
import upgradeHistory
import upLog

debug = 0

class upgradeResolver:
    """Implements the class in charge of actually running the upgrade
       algorithm based on a set of availbale informations coming from
       the installed base, the metadata source, the RHN interface,
       the policies and an history"""
    def __init__(self, base = None, RHN = None, metadata = None, \
                 sources = None, policy = None, history = None):

        if base == None:
	    base = installedBase.rpmInstalledBase()
	self.base = base

	if RHN == None:
	    RHN = rhnInterface.rhnInterface()
	self.RHN = RHN

	if metadata == None:
	    metadata = rpmMetadata.rpmMetadataSources()
	self.metadata = metadata

	if policy == None:
	    policy = upgradePolicy.upgradePolicies()
	self.policy = policy

	if history == None:
	    history = upgradeHistory.upgradeHistory()
	self.history = history

        self.initialized = 0
	self.nbUpgrades = 0
	self.failedDeps = []

        # set of headers, packages and errors related to a transaction
	self.headers = None
	self.pkgs = None
        self.errlist = None

        # all opened headers
	self.headerList = []

	# caches
	self.ccache = {}

        #
	# Those are the result of experimenting with rpmfind autoupgrade
	#
        self.maxRecursion = 9
	self.maxBatches = 10
	self.maxConflict = 20
	self.maxConflictSize = 50

	#
	# get around cached dependencies.  Gross ugly hack.
	# It's just broken in librpm, period
	#
	rpm.addMacro("_dbi_tags", 
"Packages:Name:Basenames:Group:Requirename:Providename:Triggername")
        upLog.debug("upgradeResolver created")

    #
    # Initialize the resolver:
    #   opens the installed base
    #
    # Returns 0 in case of success -1 in case of error
    #
    def init(self):
        metadb = []

        if self.base == None:
	    self.base = installedBase.rpmInstalledBase()
        if self.base == None:
	    upLog.error("upgradeResolver init failed: no base")
	    return -1
	if self.base.init() != 0:
	    upLog.error("upgradeResolver init failed: cannot open base")
	    return -1

	if self.policy == None:
	    self.policy = upgradePolicy.upgradePolicies()
	if self.policy == None:
	    upLog.error("upgradeResolver init failed: no policy")
	    return -1

	if self.history == None:
	    self.history = upgradeHistory.upgradeHistory()
	if self.history == None:
	    upLog.error("upgradeResolver init failed: no history")
	    return -1

	if self.RHN == None:
	    self.RHN = rhnInterface.rhnInterface()
	if self.RHN == None:
	    upLog.error("upgradeResolver init failed: no RHN service")
	    return -1
	if self.RHN.login() != 0:
	    upLog.error("upgradeResolver init failed: cannot login on RHN service")
	    return -1
	    

        #
	# Grab the list from the RHN server and populate the history
	#
	list = self.RHN.getList()
	if list == None:
	    upLog.error("upgradeResolver init failed: cannot get package list")
	    return -1
	    
	
	todo = 0
 	for pkg in list:
#           TODO add an arch check
#	    res = self.base.compatibleArch(pkg[4])
            priority = self.policy.checkUpgrade(pkg)

            if priority == -1:
	        continue
	    elif priority == 0:
		self.history.add(pkg, upgradeHistory.UPGRADE_STATUS_SKIPPED)
		upLog.debug("Skipping %s-%s-%s" % (pkg[0], pkg[1], pkg[2]))
		continue

	    res = self.base.isUpdate(pkg[0], pkg[1], pkg[2], pkg[3])
	    if res == 1:
		self.history.add(pkg, upgradeHistory.UPGRADE_STATUS_TODO)
		#
		# Try to find metadata sources
		#
		if string.find(pkg[0], "rpmdb") >= 0:
		    upLog.debug("New metadata source %s-%s-%s" % (pkg[0],
			      pkg[1], pkg[2]))
		    metadb.append(pkg)
	        
		upLog.debug("Update %s-%s-%s" % (pkg[0], pkg[1], pkg[2]))
		todo = todo + 1
	    elif res == 0:
		self.history.add(pkg, upgradeHistory.UPGRADE_STATUS_DONE)
	    elif res == -1:
	        if self.policy.forceCheck(pkg) == 1:
		    #
		    # Try to find metadata sources
		    #
		    if string.find(pkg[0], "rpmdb") >= 0:
			upLog.debug("New metadata source %s-%s-%s" % (pkg[0],
				  pkg[1], pkg[2]))
			metadb.append(pkg)

		    self.history.add(pkg, upgradeHistory.UPGRADE_STATUS_TODO)
		    todo = todo + 1
		    upLog.debug("Force %s-%s-%s" % (pkg[0], pkg[1], pkg[2]))
		else:
		    self.history.add(pkg, upgradeHistory.UPGRADE_STATUS_SKIPPED)

	if todo > 0:
	    upLog.progress("Found %s packages to upgrade" % todo)

        #
        # Try to install all the rpmdb updates and 
        # initialize the metadata bases
        #
	if self.metadata == None:
	    self.metadata = rpmMetadata.rpmMetadataSolver()
	if self.metadata == None:
	    upLog.error("upgradeResolver init failed: metadata source problem")
	    return -1

        for basepkg in metadb:
	    self.upgradePackage(basepkg)
        self.metadata.init()

        #
	# Find all the installed but obsoleted packages and mark the
	# associated packages for upgrade
	#
	obsol = 0
	obsoleted = self.metadata.obsoletesAll()
	for obsolete in obsoleted:
	    installed = self.base.getList(obsolete)
	    if len(installed) != 0:
		list = self.metadata.whatObsoletes(obsolete)
		for fresh in list:
		    n = fresh[0]
		    v = fresh[1]
		    r = fresh[2]
		    a = fresh[4]
		    hlist = self.history.search(n, v, r, None, a)
		    for added in hlist:
			if added[1] != upgradeHistory.UPGRADE_STATUS_TODO:
			    added[1] = upgradeHistory.UPGRADE_STATUS_TODO
			    obsol = obsol + 1

	if obsol > 0:
	    upLog.progress("Found %s packages obsoleted" % obsol)

        #
	# Sorts the upgrade list using the policy provided evaluation
	#
	f = lambda x, pol=self.policy: - pol.checkUpgrade(x)
	self.history.sort(f)
        
        self.initialized = 1
        upLog.debug("upgradeResolver initialized")
	return 0

    #
    # A log of failed deps
    #
    def logFailedDeps(self, msg = None):
        if msg == None and self.errlist == None:
            return

        if self.failedDeps == None:
	    self.failedDeps = []

        if msg == None:
	    for err in self.errlist:
		(byName, byVersion, byRelease) = err[0]
		(needsName, needsVersion) = err[1]
		flags = err[2]
		str = "%s-%s-%s requires %s " % (byName, byVersion,
		      byRelease, needsName)
		flags = flags & 0xFF
		if flags != 0 and needsVersion != None:
		    if flags & rpm.RPMSENSE_LESS:
			str = str + "<"
		    if flags & rpm.RPMSENSE_GREATER:
			str = str + ">"
		    if flags & rpm.RPMSENSE_EQUAL:
			str = str + "="
		    if flags & rpm.RPMSENSE_SERIAL:
			str = str + "S"
		    str = str + " " + needsVersion
		found = 0
		for m in self.failedDeps:
		    if m == str:
			found = 1
			break

		if found == 0:
		    self.failedDeps.insert(0, str)
	else:
	    for m in self.failedDeps:
	        if m == msg:
		    return
	    self.failedDeps.insert(0, msg)

    #
    # The header cache lookup
    #
    def getHeader(self, pkg):
        if pkg == None:
	    return None

        for e in self.headerList:
	    p = e[0]
	    h = e[1]
	    if p is pkg:
	        return h
	    if pkg[0] == p[0] and pkg[1] == p[1] and pkg[2] == p[2] and \
	       pkg[3] == p[3] and pkg[4] == p[4]:
	        return h

	try:
	    h = self.RHN.getHeader(pkg)
	except IOError:
	    upLog.error("IOError:", sys.exc_type)
	    h = None
	except OSError:
	    upLog.error("OSError:", sys.exc_type)
	    h = None

	if h != None:
	    self.headerList.append((pkg, h))

        return h
	
    #
    # The header cache cleanup
    #
    def delHeader(self, pkg):
        if pkg == None:
	    return None

        for e in self.headerList:
	    p = e[0]
	    if p is pkg:
	        self.headerList.remove(e)
		return
	    if pkg[0] == p[0] and pkg[1] == p[1] and pkg[2] == p[2] and \
	       pkg[3] == p[3] and pkg[4] == p[4]:
	        self.headerList.remove(e)
		return
	        
	
    #
    # Generic routine to search if a header provide a dependancy
    # either by package name, NVR, provides or filename
    #
    def headerProvides(self, h, name, flags = 0, ver = None, rel = None):
        if h == None or name == None:
	    return 0

	str = "check %s-%s-%s headerProvides %s " % (h['name'],
	      h['version'], h['release'], name)
	if flags != 0 and ver != None:
	    if flags & rpm.RPMSENSE_LESS:
		str = str + "<"
	    if flags & rpm.RPMSENSE_GREATER:
		str = str + ">"
	    if flags & rpm.RPMSENSE_EQUAL:
		str = str + "="
	    if flags & rpm.RPMSENSE_SERIAL:
		str = str + "S"
	    str = str + ver
	upLog.debug(str, 1)

        ret = 0
	if name == h['name']:
	    if flags == 0 or ver == None or ver == "":
	        upLog.debug("   okay",1)
	        ret = 1
	    else:
		version = h['version']
		release = h['release']
		r = rel
		if r == None or r == '':
		    r = release
	     
		cmp = rpm.labelCompare(('', version, release), ('', ver, r))
		upLog.debug("   cmp = %s" % cmp,1)

		if flags & rpm.RPMSENSE_LESS and cmp < 0:
		    upLog.debug("   NVR okay",1)
		    ret = 1
		elif flags & rpm.RPMSENSE_GREATER and cmp > 0:
		    upLog.debug("   NVR okay",1)
		    ret = 1
		elif flags & rpm.RPMSENSE_EQUAL and cmp == 0:
		    upLog.debug("   NVR okay",1)
		    ret = 1
	elif name[0] == '/':
	    files = h['filenames']
	    if files:
		for x in files:
		    if x == name:
			upLog.debug("   file okay",1)
			ret = 1
			break
	    upLog.debug("   failed: does not provide the file",1)
	elif flags == 0 or ver == None or ver == "":
	    provides = h['provides']
	    if provides:
		for x in provides:
		    if x == name:
			upLog.debug("   provide okay",1)
			ret = 1
			break
	    if ret == 0:
		obsoletes = h['obsoletes']
		if obsoletes:
		    for x in obsoletes:
			if x == name:
			    upLog.debug("   obsoletes okay",1)
			    ret = 1
			    break

	    if ret == 0:
		upLog.debug("   failed: does not provide or obsolete it", 1)

	return ret

    #
    # Add a package to the transaction set, it will try first
    # to install/upgrade it standalone
    # 
    #
    def addPackage(self, pkg, header = None):
        if pkg == None:
	    return -1

	upLog.debug("Adding package %s-%s-%s" % (pkg[0], pkg[1], pkg[2]))

        if header == None:
	    header = self.getHeader(pkg)
	    if header == None:
		upLog.error("failed to get %s-%s-%s header !" % (
		      pkg[0], pkg[1], pkg[2]))
		self.history.update(pkg, upgradeHistory.UPGRADE_STATUS_BROKEN)
		return -1

	self.history.update(pkg, upgradeHistory.UPGRADE_STATUS_TODO)

        #
	# Try first to see if this installs standalone
	# Goal is to cut leaves from the tree it really simplifies things
	#
	code, list = self.testTransaction([header], [pkg])
	if code < 0:
	    return 0
	if code == 1:
	    return 1
        
	if self.headers == None or self.pkgs == None:
	    if self.headers != None or self.pkgs != None:
		upLog.error("addPackage: internal error, size mismatches")
		return (-1)
	    self.headers = [ header ]
	    self.pkgs = [ pkg ]
	    if self.errlist == None:
	        self.errlist = list
	else:
	    if self.headers == None or self.pkgs == None:
		upLog.error("addPackage: internal error, size mismatches")
		return (-1)
	    if len(self.headers) != len(self.pkgs):
		upLog.error("addPackage: internal error, size mismatches")
		return (-1, None)
	    self.headers.append(header)
	    self.pkgs.append(pkg)

	return 0

    #
    # Add a pkg/header to the current transaction:
    #
    def addTransaction(self, p, h):
        if p == None:
	    return -1
        if h == None:
	    return -1

	# Check first that a package providing it is not installed now
	if self.base.isInstalled(p) == 1:
	    upLog.debug("addTransaction: %s-%s-%s already installed" % (
		      p[0], p[1], p[2]))
	    return 0

	for pkg in self.pkgs:
	    if pkg is p:
		return 0
	for header in self.headers:
	    if header is h:
		return 0

        self.addPackage(p, h)

	return 1

    #
    # The header cache search
    #
    def searchHeader(self, name, flags = 0, ver = None, rel = None):
        if name == None:
	    return None

        ret = []

        for e in self.headerList:
	    h = e[1]
	    if self.headerProvides(h, name, flags, ver, rel):
	       ret.append(e)

        upLog.debug("searchHeader: %s" % (ret),1)

        return ret

    #
    # Try to find a package providing a given missing dependancy.
    # We first look in the history list, since this probably available
    # as an update too, then we do a strict lookup in the general
    # rpmfind database. If we find the corresponding package it
    # is added in the upgrade list.
    #
    # Returns 1 in case of success, 0 if lookup failed, -1 in case of error
    #
    def lookupDep(self, name, flags = 0, ver = None, rel = None, local = 0):
        if name == None:
	    return -1

	str = "lookupDep %s " % name 
	if flags != 0 and ver != None:
	    if flags & rpm.RPMSENSE_LESS:
		str = str + "<"
	    if flags & rpm.RPMSENSE_GREATER:
		str = str + ">"
	    if flags & rpm.RPMSENSE_EQUAL:
		str = str + "="
	    if flags & rpm.RPMSENSE_SERIAL:
		str = str + "S"
	    str = str + ver
	upLog.debug(str, 1)
	    
	# Check first that the policy allows to upgrade it
	ret = self.policy.checkUpgrade((name, "", "", "", "", ""))
	if ret <= 0:
	    if ret == 0:
	        self.logFailedDeps("Upgrade of %s is forbidden" % name)
		return 0
	    if ret < 0:
	        self.logFailedDeps("Upgrade of %s is an exception" % name)
		return 0

	# Check that a package providing it was not turned off
	lst = self.searchHeader(name, flags, ver, rel)
	if lst != None and len(lst) > 0:
	    if len(lst) > 1:
	        upLog.warning("lookupDep found %s packages for %s" %
		              ((len(lst), name)))
	    for e in lst:
		p = e[0]
		h = e[1]
		self.addTransaction(p, h)
		return 1
	    
        # Then lookup in the metadata base for packages providing it
	if self.metadata != None:
	    list = self.metadata.whatProvides(name)
	    for pkg in list:
		n = pkg[0]
		v = pkg[1]
		r = pkg[2]
		e = pkg[3]
		a = pkg[4]
		upLog.debug("metadada: %s-%s-%s provides %s" % (n, v, r, name))
		lst = self.history.search(n, v, r, e, a)
		if len(lst) == 0:
		    upLog.debug(" %s-%s-%s not found in history" % (n, v, r))
		    lst = self.history.search(n, v, None, None, a)
		if len(lst) == 0:
		    upLog.debug(" %s-%s not found in history" % (n, v))
		    lst = self.history.search(n, None, None, None, a)
		if len(lst) == 0:
		    upLog.debug(" %s not found in history" % (n))

		for e in lst:
		    p = e[0]
		    h = self.getHeader(p)
		    if h != None:
			upLog.debug(" checking %s-%s-%s header" % (n, v, r))
			if self.headerProvides(h, name, flags, ver, rel):
			    self.addTransaction(p, h)
			    return 1
			else:
			    upLog.debug(" failed !")
			    
	    else:
	        upLog.debug("found no provides for %s in metadata" % (name))

        # Then lookup in the metadata base for packages obsoleting it
	if self.metadata != None:
	    list = self.metadata.whatObsoletes(name)
	    for pkg in list:
		n = pkg[0]
		v = pkg[1]
		r = pkg[2]
		e = pkg[3]
		a = pkg[4]
		upLog.debug("metadada: %s-%s-%s provides %s" % (n, v, r, name))
		lst = self.history.search(n, v, r, e, a)
		if len(lst) == 0:
		    upLog.debug(" %s-%s-%s not found in history" % (n, v, r))
		    lst = self.history.search(n, v, None, None, a)
		if len(lst) == 0:
		    upLog.debug(" %s-%s not found in history" % (n, v))
		    lst = self.history.search(n, None, None, None, a)
		if len(lst) == 0:
		    upLog.debug(" %s not found in history" % (n))

		for e in lst:
		    p = e[0]
		    h = self.getHeader(p)
		    if h != None:
			upLog.debug(" checking %s-%s-%s header" % (n, v, r))
			if self.headerProvides(h, name, flags, ver, rel):
			    self.addTransaction(p, h)
			    return 1
			else:
			    upLog.debug(" failed !")
			    
	    else:
	        upLog.debug("found no obsoletes for %s in metadata" % (name))

        # TODO: search RHN
	upLog.debug("lookupDep: local failed, TODO: search RHN")
	
        return 0

    #
    # Try to find a package providing compatibility with an older library
    #
    # Returns 1 in case of success, 0 if lookup failed, -1 in case of error
    #
    def lookupCompat(self, name):

	ret = self.lookupDep(name, 0, None, None, 1)
	if ret > 0:
	    return ret
        #
	# TODO: autoupgradeLookupCompat from autoupgrade.c
	#
	upLog.debug("TODO: lookupCompat(%s)" % name)
        return 0

    #
    # Try a transaction make of an header list, it will handle
    #
    # Returns an error code/package number and an error list in case
    #         the precheck fails
    #
    def testTransaction(self, headers, pkgs, force = 0):
        if headers == None:
	    upLog.warning("testTransaction: no headers")
	    return (-1, None)
        if pkgs == None:
	    upLog.warning("testTransaction: no pkgs")
	    return (-1, None)
	if len(headers) != len(pkgs):
	    upLog.warning("testTransaction: internal error, size mismatches")
	    return (-1, None)

        #
	# Verify that the packages are not installed
	#

        for pkg in pkgs:
	    if self.base.isInstalled(pkg):
	        upLog.debug("package %s-%s-%s is already installed" % (pkg[0],
		          pkg[1],pkg[2]))
		ret = self.history.update(pkg,
		                   upgradeHistory.UPGRADE_STATUS_DONE)
		found = 0
		for header in headers:
		    n = header['name']
		    v = header['version']
		    r = header['release']
		    if pkg[0] == n and pkg[1] == v and pkg[2] == r:
		        headers.remove(header)
			pkgs.remove(pkg)
			found = 1
			break
		if found == 0:
		    upLog.warning("package %s-%s-%s not found in headers", (pkg[0],
		          pkg[1],pkg[2]))
		    return(-1, None)
		           
        if len(headers) == 0:
	    return(0, None)

        #
	# If one really need to bypass rpmCheck
	#
        if force == 0:
	    ret = self.base.checkTransaction(headers);
	    if ret != None:
		return (0, ret)
	    upLog.debug("transaction with %s packages okay" % len(headers))
	else:
	    upLog.debug("transaction with %s packages forced" % len(headers))
	
	for pkg in pkgs:
	    upLog.debug("fetching package: %s-%s-%s.%s" % (pkg[0],
		      pkg[1],pkg[2],pkg[4]),1)
	    try:
		ret = self.RHN.getPackage(pkg)
	    except:
		upLog.error("Failed to fetch package: %s-%s-%s.%s" % (pkg[0],
		      pkg[1],pkg[2],pkg[4]))
		self.history.update(pkg, upgradeHistory.UPGRADE_STATUS_BROKEN)
		return (-1, None)

	ret = self.base.runTransaction(headers,
				       self.RHN.storageDirGet())
			  # TODO: self.RHN.installCallbackGet())

	if ret == None:
	    #
	    # Post processing of installed packages
	    #
	    for pkg in pkgs:
	        self.nbUpgrades = self.nbUpgrades + 1
		ret = self.history.update(pkg,
		                   upgradeHistory.UPGRADE_STATUS_DONE)
		if ret == None:
		    upLog.warning("%s-%s-%s not found in history !" % (pkg[0],
		          pkg[1], pkg[2]))

		scripts = self.policy.postScript(pkg)
		for script in scripts:
		    upLog.progress("Running post script for %s: %s" % (pkg[0], script))
		    try:
			exec(script)
		    except:
			upLog.error("Post script '%s' generated an exception" % script)

		# If it's a metadata base invalidate them.
		if string.find(pkg[0], "rpmdb") >= 0:
		    self.metadata.invalidate()

		# Remove it from the local cache
		self.delHeader(pkg)

            #
	    # Remove installed packages and headers from the current
	    # transaction lists
	    #
	    if self.pkgs != None:
		for pkg in pkgs:
		    for p in self.pkgs:
			if p is pkg:
			    self.pkgs.remove(pkg)

	    if self.headers != None:
		for header in headers:
		    for h in self.headers:
			if h is header:
			    self.headers.remove(header)

	    upLog.debug("runTransaction Okay")
	else:
	    upLog.error("runTransaction pbm:", ret)

	return (len(pkgs), None)
	

    #
    # Find the package who generated a problem in the transaction list
    #
    # Returns the number of the packages in the list or -1 if not found
    #
    def findConflict(self, byName, byVersion, byRelease):
        if self.pkgs == None:
	    return -1
	i = 0
	while i < len(self.pkgs):
	    if self.pkgs[i][0] == byName and self.pkgs[i][1] == byVersion and \
	       self.pkgs[i][2] == byRelease:
	       return i
	    i = i + 1

	return -1
        
    #
    # Try to remove a package blocking the upgrade for which no update
    # has been found and no compatibility library required was found either
    # Check the policy manager first
    #
    # Returns 1 in case of success, 0 if failed, -1 in case of error
    #
    def tryRemove(self, name, version, release):
        #
	# TODO: autoupgradeTryRemove from autoupgrade.c
	#
	upLog.warning("TODO: tryRemove(%s-%s-%s)" % (name, version, release))
        return 0


    #
    # Try to solve the conflict list and end up with a completing
    # transaction
    # Equivalent of the 2nd part of autoupgradeSolveConflicts()
    # in rpmfind autoupgrade.c code
    #
    # Returns the number of packages upgraded at that step or -1 in case of
    #         error
    #
    def solveConflicts(self, level = 0, islast = 0):
        if self.errlist == None:
            return 0

        if level == self.maxRecursion:
            upLog.debug("Conflict too deep: aborting",1)
	    self.logFailedDeps()
            return 0

	# First pass over the error list, trying to find direct conflicts
	# from the upgrade packages or missing "compatibility" packages
        progress = 0
	numFixes = 0
	sharedLibUpgrade = 0
	sharedLibPbm = []
	#nbPackages = len(self.pkgs)

	fixed = map(lambda x : 0, self.errlist)

        index = -1
        for err in self.errlist:
	    index = index + 1
	    (byName, byVersion, byRelease) = err[0]
	    (needsName, needsVersion) = err[1]
	    flags = err[2]
	    sense = err[4]
	    #
	    # A bit of processing is needed:
	    #
	    flags = flags & 0xFF
	    needsRelease = None
	    if needsVersion != None and string.find(needsVersion, '-') > 0:
	        l = string.split(needsVersion, '-')
		needsVersion = l[0]
		needsRelease = l[1]

	    l = string.split(needsVersion)

            culprit = self.findConflict(byName, byVersion, byRelease)
            if culprit >= 0:
		# I] try to handle cases where the new package introduced
		# direct new dependancies/conflicts
                upLog.debug("Error generated by %s-%s-%s" % (byName,
                                           byVersion, byRelease),1)

                if sense == rpm.RPMDEP_SENSE_REQUIRES:
                    # The new package requires another one
                    # We try to do a closure either by scanning
                    # history about this dependancy or doing a
                    # metadata lookup
                    upLog.debug("  New package %s requires dep %s" % (byName,
                                needsName),1)
		    ret = self.lookupDep(needsName, flags, needsVersion,
		                         needsRelease)
		    if ret > 0:
		        progress = 1
			numFixes = numFixes + 1
			fixed[index] = 1
			continue;

		    # If the lookup failed and we are within a 
		    # resolution tree simply drop the package
		    if level == self.maxRecursion -1:
			upLog.debug("  Failed, drop %s from list" % byName,1)
			del self.pkgs[culprit]
			del self.headers[culprit]
		elif sense == rpm.RPMDEP_SENSE_CONFLICTS:
		    # The new package is in conflict with another
		    # (installed) one.
                    upLog.debug("  New package %s conflicts with %s" % (byName,
                                needsName),1)
		    ret = self.lookupDep(needsName, 0, None, None)
		    if ret > 0:
		        progress = 1
			numFixes = numFixes + 1
			fixed[index] = 1
			continue;

            else: # culprit < 0 i.e one of the installed packages raised it
		# II] This is one of the installed packages which would
		# break due to the update, one of the replaced packages
		# provides a ressource that is required by an installed
		# packages. Try to bring compatibilities updates
		if string.find(needsName, ".so") > 0:
		    # This is a direct dependancy on a library
		    # try to find first the compatibility lib
		    upLog.debug("old package %s requires %s checking compat" % (
			      byName, needsName),1)
		    # Try first to bring a compatibility library
		    ret = self.lookupCompat(needsName)
		    if ret > 0:
			# TODO: double check the original algorithm seems
			#       an extra step was removed
			#  providePackageNVR equiv from python ????
			fixed[index] = 1
			numFixes = numFixes + 1
			progress = 1
			sharedLibUpgrade = sharedLibUpgrade + 1
		    else:
		        sharedLibPbm.append(needsName)

	# If a compatibility library was found try to rerun the
	# transaction, cut here to avoid bringing in too many packages
	# at once.
	if progress > 0 and len(sharedLibPbm) == 0:
	    code, list = self.testTransaction(self.headers, self.pkgs)
	    if code > 0:
		self.headers = None
		self.pkgs = None
		self.errlist = None
		return code
	    if code < 0:
	        self.headers = None
		self.pkgs = None
		self.errlist = None
		return code

	    self.errlist = list
	    code = self.solveConflicts(level + 1, islast)
	    return code

	# III]
	# Second pass over the error list
	# if no compatibility library was found then try to update
	# the complaining packages.
	index = -1
        for err in self.errlist:
	    index = index + 1
	    if fixed[index] == 1:
	        continue
	    (byName, byVersion, byRelease) = err[0]
	    (needsName, needsVersion) = err[1]
	    flags = err[2]
	    sense = err[4]
	    #
	    # A bit of processing is needed:
	    #
	    flags = flags & 0xFF
	    needsRelease = None
	    if needsVersion != None and string.find(needsVersion, '-') > 0:
	        l = string.split(needsVersion, '-')
		needsVersion = l[0]
		needsRelease = l[1]

            culprit = self.findConflict(byName, byVersion, byRelease)
            if culprit < 0:
		# This is one of the installed packages which would
		# break due to the update, one of the replaced packages
		# provides a ressource that is required by an installed
		# packages.
                upLog.debug("compat %s not found checking %s upgrade" % (
		            needsName, byName),1)
		ret = self.lookupDep(byName, 0, None)
		if ret > 0:
		    progress = 1
		    numFixes = numFixes + 1
		    fixed[index] = 1
		else:
		    upLog.warning("DEADLOCK: caused by installed package %s-%s-%s" % (
		          byName, byVersion, byRelease))
		    if self.tryRemove(byName, byVersion, byRelease) > 0:
			progress = 1
			numFixes = numFixes + 1
			fixed[index] = 1

        upLog.debug("%s conflicts, %s fixes found" % (len(self.errlist),
	            numFixes),1)

	code, list = self.testTransaction(self.headers, self.pkgs)

	if code > 0:
	    self.headers = None
	    self.pkgs = None
	    self.errlist = None
	    return code
	if code < 0:
	    self.headers = None
	    self.pkgs = None
	    self.errlist = None
	    return code
        self.errlist = list
        code = self.solveConflicts(level + 1, islast)
        return code

    #
    # Try to upgrade a given package, this ends up trying to
    # find a transaction including this package which will run
    # without errors.
    #
    # Returns the number of packages upgraded at that step or -1 in case of
    #         error
    #
    def upgradePackage(self, pkg):
        #
	# We restart the transaction from scratch.
	# TODO: keep the headers Hopefully they are cached
	#
	self.headers = None
	self.pkgs = None
	self.errlist = None

        ret = self.addPackage(pkg)
	if ret == -1:
	    return -1
	if ret == 1:
	    return 1

        self.solveConflicts(0, 0)
        
	return 0

    #
    # Start the upgrader
    #
    # Returns the number of packages upgraded
    #
    def upgrade(self):
	import time
	start = time.time()

        if self.initialized == 0:
	    self.init()
	    if self.initialized == 0:
	        return -1
	
	upLog.debug("Starting the upgrade algorithm")

        #
	# Try installing some of the packages from the history
	# while there is progresses made at each scan
	#
        progress = 1
	left = 0
	while progress > 0:
	    progress = 0
	    no = 0
	    left = 0
	    nbUpgrades = self.nbUpgrades
	    while no < self.history.len():
		upgrade = self.history.item(no)
		no = no + 1

		if upgrade[1] == upgradeHistory.UPGRADE_STATUS_TODO:
		    left = left + 1
		    pkg = upgrade[0]
		    self.upgradePackage(pkg)
	    progress = self.nbUpgrades - nbUpgrades

        
	if self.nbUpgrades > 0:
	    delta = time.time() - start
	    upLog.progress("------")
	    upLog.progress("Upgraded %d packages in %d:%d" % (self.nbUpgrades,
	          delta / 60,  delta % 60))
        if left > 0:
	    upLog.progress("------")
	    upLog.progress("%s packages not upgraded left:" % left)
	    no = 0
	    while no < self.history.len():
		upgrade = self.history.item(no)
		no = no + 1
		if upgrade[1] == upgradeHistory.UPGRADE_STATUS_TODO:
		    pkg = upgrade[0]
		    upLog.progress("   %s-%s-%s.%s.rpm" % (pkg[0], pkg[1],
		          pkg[2], pkg[4]))
	    upLog.progress("%s unresolvable conflicts found:" % len(self.failedDeps))
	    for msg in self.failedDeps:
	        upLog.progress("   %s" % msg)


    #
    # Stop the resolver if running
    #
    def stop(self):
        upLog.warning("Not implemented")


#
# Small test routine to do basic tests
#
def main():
    global debug
    solver = None

    args = sys.argv[1:]
    while args:
        if args[0] == '--init' or args[0] == '-init':
	    if args[0] == None:
		return

	    if solver == None:
		solver = upgradeResolver()
	    if solver == None:
		sys.exit(-1)

            solver.init()

        if args[0] == '--upgrade' or args[0] == '-upgrade':
	    if args[0] == None:
		return

	    if solver == None:
		solver = upgradeResolver()
	    if solver == None:
		sys.exit(-1)

            solver.upgrade()

        elif args[0] == '--verbose' or args[0] == '-v':
	    debug = debug + 1
            rhnInterface.debug = rhnInterface.debug + 1
            rpmMetadata.debug = rpmMetadata.debug + 1
            upgradePolicy.debug = upgradePolicy.debug + 1
	    if  debug > 1:
		upgradeHistory.debug = upgradeHistory.debug + 1
		installedBase.debug = installedBase.debug + 1

        else:
            sys.stderr.write('unknown argument '+ args[0] + '\n')
	    sys.stderr.write('use --init or -v\n')
	    sys.exit(-1)
        args = args[1:]


if __name__ == "__main__":
    main()
