#!/usr/bin/python

# Copyright (c) 1999-2002 Red Hat, Inc.  Distributed under GPL.
#
# Author: Adrian Likins <alikins@redhat.com>

# the idea here is to implement something that loooks like
# a rpm.transactionSet, including .depcheck() but instead of
# letting rpm do it the hardway (loading headers, grinding out
# deps, sig/header checks, file fingerprints, etc, etc) try
# to see if we can come up with a "close enough" approach
# that might be faster. Aka, the "apt" way

import sys
import os
import string
import shelve

import rpm

import up2dateUtils
import config

class RpmClass:
    def __init__(self, hdr=None, filename=None):
        self.requires = {}
        self.provides = {}
        self.obsoletes = {}
        self.conflcits = {}
        # these can be big, hash lookup probabaly faster
        # than iterating a list
        self.files = {}
        if hdr:
            self.populateWithHdr(hdr)

        if filename:
            ts = rpm.TransactionSet()
            fd = os.open(filename, os.O_RDONLY)
            hdr = ts.hdrFromFdno(fd)
            os.close(fd)
        self.populateWithHdr(hdr)

    def getRequires(self):
        list  = []
        for key in self.requires.keys():
            for req in self.requires[key]:
                list.append((key, req[0],req[1]))
        return list

    def getProvides(self):
        list  = []
        for key in self.provides.keys():
            for prov in self.provides[key]:
                list.append((key, prov[0], prov[1]))
        return list

    def getObsoletes(self):
        list  = []
        for key in self.obsoletes.keys():
            for prov in self.obsoletes[key]:
                list.append((key, prov[0], prov[1]))
        return list

    def getConflicts(self):
        list  = []
        for key in self.conflicts.keys():
            for prov in self.conflicts[key]:
                list.append((key, prov[0], prov[1]))
        return list

    def getFiles(self):
        list = self.files.keys()
        list.sort()
        return list

    def getNVRE(self):
        return (self.name, self.version, self.release, self.epoch)

    def getNVREA(self):
        return (self.name, self.version, self.release, self.epoch, self.arch)

    def getNVR(self):
        return (self.name, self.version, self.release)

    def populateWithHdr(self, hdr):
        self.requires = joinDepInfoList(hdr, "require")
        self.provides = joinDepInfoList(hdr, "provide")
        self.obsoletes = joinDepInfoList(hdr, "obsolete")
        self.conflicts = joinDepInfoList(hdr, "conflict")
        filelist = hdr['filenames']
        self.files = filelist

        self.name = hdr['name']
        self.version = hdr['version']
        self.release = hdr['release']
        self.epoch = hdr['epoch']
        self.arch = hdr['arch']

        del hdr


def joinDepInfoList(hdr, deptype):
    tmpdict = {}
    
    names = "%sname" % deptype
    flags = "%sflags" % deptype
    versions = "%sversion" % deptype

    depnames = hdr[names]
    depflags = hdr[flags]
    depversions = hdr[versions]

#    print depnames
#    print depflags
#    print depversions
    if not depnames or not len(depnames):
        return tmpdict

    # special case
    if len(depnames) == 1 and (deptype == "conflict" or deptype == "obsolete" or deptype == "provide"): 
        tmpdict[depnames[0]] = [(depflags, depversions[0])]
        return tmpdict

    for i in range(len(depnames)):
        name = depnames[i]
        flag = depflags[i]
        version = depversions[i]
        if not tmpdict.has_key(name):
            tmpdict[name] = []
        tmpdict[name].append((flag, version))

    return tmpdict


class VTransaction:
    def __init__(self):
        # key being the "name", and value a list of cases, maybe sorted...
        self.providesCache = {}
        self.requires = {}
        self.obsoletes = {}
        self.conficts = {}
        self.provides = {}

#        self.pkgToRequires = {}
#        self.pkgToProvides = {}
#        self.pkgToObsoletes = {}
        self.pkgToConflicts = {}

        # any give req/pro/obs/conflict can
        # show up in more than one package
        # key is tuple, value is a list of NVREA tuples
        cfg = config.initUp2dateConfig()
#        self.requireToPkgs = shelve.open("%s/requiretopkg.shelve" % cfg['storageDir'], flag="n")
#        self.provideToPkgs = shelve.open("%s/providetopkg.shelve" % cfg['storageDir'], flag="n")
        self.requireToPkgs = {}
        self.provideToPkgs = {}
        self.provideNameToProvideTuple = {}
        self.solvedRequires = {}

        # this takes huge amount of ram, so use a shelve
        self.filenameIndex = {}
#        self.filenameIndex = shelve.open("%s/filenameindex.shelve" % cfg['storageDir'], flag="n")


        
        self.obsoleteToPkgs = {}
        self.conflictToPkgs = {}
        
        # so we know what to ask for
        self.unsolvedRequires = {}

        # we need a graph of what package depends on
        # what packages if we want to be able to
        # split stuff up
        # this is a dict, key is pkg that depends on value
        self.depGraph = {}

        self.ts = rpm.TransactionSet()

    def check():
        pass

    def order():
        pass

    def addErase():
        pass


    def _mergeDicts(self, orig, new):
        for key in new.keys():
            if not orig.has_key(key):
                orig[key] = []
            # dont add it if we already have it
            if new[key] not in orig[key]:
                orig[key].append(new[key])
        return orig


#    def mergeDicts(orig, newlist):
#        for item in newlist:
#            if not orig
    
    def addInstall(self, header, header2, instMethod):
        # take a header, and install/upgrade
        name = header['name']
        rpmobj = RpmClass(hdr=header)


#        self.pkgToRequires[rpmobj.getNVREA()] = rpmobj.getRequires()
#        self.pkgToProvides[rpmobj.getNVREA()] = rpmobj.getProvides()
#        for provideTuple in self.pkgToProvides.keys():
#            self.provideNameToProvideTuple[provideTuple[0]] = provideTuple
#        self.pkgToObsoletes[rpmobj.getNVREA()] = rpmobj.getObsoletes()
        self.pkgToConflicts[rpmobj.getNVREA()] = rpmobj.getConflicts()

        for require in rpmobj.getRequires():
            if not self.requireToPkgs.has_key(require):
                self.requireToPkgs[require] = []
            self.requireToPkgs[require].append(rpmobj.getNVREA())

        for provide in rpmobj.getProvides():
            if not self.provideToPkgs.has_key(provide):
                self.provideToPkgs[provide] = []
            self.provideToPkgs[provide].append(rpmobj.getNVREA())

        # FIXME: guess what kiddies, more than one thing can provide something
        # and guess what? some packages provide themselves more than once! with
        # different versions even! (kernel-drm and perl being two good examples)
        for provideTuple in self.provideToPkgs.keys():
            if not self.provideNameToProvideTuple.has_key(provideTuple[0]):
                self.provideNameToProvideTuple[provideTuple[0]] = []
            self.provideNameToProvideTuple[provideTuple[0]].append(provideTuple)

#        # populate a file list...
        for filename in rpmobj.files:
            # files dont have versions
            # we dont really need the provide info since
            # theres no sense or version, so dont wast the mement
            self.filenameIndex[filename] = 1

        
        for obsolete in rpmobj.getObsoletes():
            if not self.obsoleteToPkgs.has_key(obsolete):
                self.obsoleteToPkgs[obsolete] = []
            self.obsoleteToPkgs[obsolete].append(rpmobj.getNVREA())

        for conflict in rpmobj.getConflicts():
            if not self.conflictToPkgs.has_key(conflict):
                self.conflictToPkgs[conflict] = []
            self.conflictToPkgs[conflict].append(rpmobj.getNVREA())

        # merge the info from the latest header into the dict
        # representing the ts
#        self.requires = self._mergeDicts(self.requires, requiresdict)
#        self.provides = self._mergeDicts(self.provides, providesdict)
#        self.obsoletes = self._mergeDicts(self.obsoletes, obsoletedict)
#        self.conflicts = self._mergeDicts(self.conflicts, conflicts)



    def run(self,arg1, arg2):
        print "run called with: %s %s" % (arg1, arg2)

    def findConflicts(self):
        conflicts = []
        for conflict in self.conflictToPkgs.keys():
            # any list of pkgs will suffice
            # FIXME: need to check for conflicts with installed
            # packages as well
            if not self.pkgToConflicts.has_key(conflict[0]):
                continue
            conflicts.append(conflict)

        return conflicts

    def checkDepMatch(self):
        pass


    def parseVersion(self,ver):
        s = ver
        epoch = ""
        if string.find(s, ":") >= 0:
            arr = string.split(s, ":", 1)
            epoch = arr[0]
            s = arr[1]
        release = "0"
        if string.find(s, '-') >= 0:
            arr = string.split(s, "-", 1)
            release = arr[1]
            s = arr[0]
        return s, release, epoch


    def _checkEqual(self, require,pv, pr, pe ):
        if require[2] == "%s" % pv:
            return 1
        if require[2] == "%s-%s" % (pv, pr):
            return 1
        if pe:
            if require[2] == "%s:%s" % (pe, pv):
                #                        print "====== %s satisified with %s" % (require, provtuple)
                return 1
            if require[2] == "%s:%s-%s" % (pe, pv, pr):
                #                        print "====== %s satisified with %s" % (require, provtuple)
                return 1
        return 0

    def _checkGreater(self, v, r, e, pv, pr, pe):
        foo1 = ('a', v,r,e)
        foo2 = ('a', pv, pr, pe)
        ret = up2dateUtils.comparePackages(foo2, foo1)
        if ret > 0:
            return 1
        return 0

    def _checkLess(self, v, r, e, pv, pr, pe):
         foo1 = ('a', v,r,e)
         foo2 = ('a', pv, pr, pe)
         ret = up2dateUtils.comparePackages(foo2, foo1)
         if ret < 0:
             return 1
         return 0

    def verifyDepSolve(self, require, provtuple):
         sense = require[1]
         v,r,e = self.parseVersion(require[2])
         pv,pr,pe = self.parseVersion(provtuple[2])
         if e == "":
             e = None
         if pe == "":
             pe = None

         if (sense == 0):
             #                    print "-------- no version specified for %s so solved with %s" % (require, provtuple)
             return 1

         if (sense & rpm.RPMSENSE_PREREQ):
             if (sense == rpm.RPMSENSE_PREREQ):
                 return 1
             # we hit on prereq and there are no other conditions or vercomps to do
             if not (sense & rpm.RPMSENSE_EQUAL) and not (sense & rpm.RPMSENSE_GREATER) \
                    and not (sense & rpm.RPMSENSE_LESS):
                 return  1
             
         if (sense & rpm.RPMSENSE_FIND_REQUIRES):
             if (sense == rpm.RPMSENSE_FIND_REQUIRES):
                 return 1

         # equal or equal or greater
         if (sense & rpm.RPMSENSE_EQUAL):
             if self._checkEqual(require, pv, pr, pe):
                 return 1

         # greater than or greater or equal
         if (sense & rpm.RPMSENSE_GREATER):
             if self._checkGreater(v,r,e,pv,pr,pe):
                 return 1

         # less seems very uncommon so do it last
         if (sense & rpm.RPMSENSE_LESS):
             if self._checkLess(v, r, e, pv, pr, pe):
                 return 1

         if (sense & rpm.RPMSENSE_CONFIG):
##                    print "%s & rpm.RPMSENSE_EQUAL" % (sense & rpm.RPMSENSE_EQUAL)
##                    print "%s & rpm.RPMSENSE_GREATER" % (sense & rpm.RPMSENSE_GREATER)
##                    print "%s & rpm.RPMSENSE_PREREQ" % (sense & rpm.RPMSENSE_PREREQ)
##                    print "%s & rpm.RPMSENSE_FIND_REQUIRES" % (sense & rpm.RPMSENSE_FIND_REQUIRES)
             if sense == rpm.RPMSENSE_CONFIG:
                 #                        print "config stuff: %s" % (require,)
                 return 1 

             if ( sense & rpm.RPMSENSE_EQUAL):
                 if self._checkEqual(require, pv, pr, pe):
                     return 1
         return 0
         
    # just for testing, but very simple first pass at computing
    # transaction closure
    def findUnsolvedRequires(self):
        bar = []
        unsolved = []
        # FIXME: since this is multistage, we should keep tracking
        # of the already solved deps and not recompute them
        for require in self.requireToPkgs.keys():
            if self.solvedRequires.has_key(require):
                #print "cache hit on %s" % (require,)
                continue
            #print require
            if self.provideNameToProvideTuple.has_key(require[0]):
                for provtuple in self.provideNameToProvideTuple[require[0]]:
                    if self.verifyDepSolve(require, provtuple):
                        self.solvedRequires[require] = 1
                        break
                continue
    #                for pkg in self.provieToPkgs[require]:
    #                    # we can multiple packages providing something in
    #                    # theory...
    #                    for provpkg in self.provideToPkgs[require]:
#                        self.depGraph[pkg] = self.provideToPkgs

            else:
                sense = require[1]
                if (sense & rpm.RPMSENSE_RPMLIB):
                    if string.find(require[0], "rpmlib(") == 0:
                        continue

                # See if it is a filename:
                if self.filenameIndex.has_key(require[0]):
                    # FIXME, do lookup to packagename so we can collapse
                    # into packagename for the dep graph
                    self.solvedRequires[require] = 1
                    continue

# FIXME: this need to be smart enough to do ver checks and such as well,
# maybe refactor the above
            if self.findDepLocal(require):
                #print "db hit on: %s" % (require,)
                self.solvedRequires[require] = 1
                continue
            unsolved.append(require)
        return unsolved

    def order(self):
        print "called order"

    def check(self):
        deplist = []
        unsolvedReqs = self.findUnsolvedRequires()
        conflicts = self.findConflicts()
        #print "conflicts: %s" % conflicts
        for item in unsolvedReqs:
            pkgs = self.requireToPkgs[item]
            for pkg in pkgs:
                deplist.append(((pkg[0], pkg[1], pkg[2]),
                               (item[0], item[2]),
                                item[1], "",
                                rpm.RPMDEP_SENSE_REQUIRES))

        return deplist
        
    def setFlags(self,flags):
        pass

    def setVSFlags(self,flags):
        pass

    def setProbFilter(self, flags):
        pass


    def findDepLocal(self, dep):
        if self.providesCache.has_key(dep):
            #print "CACHE hit: %s" % (dep,)
            return 1
        name = dep[0]
        flag = dep[1]
        version = dep[1]
        if name[0] == '/':
            tagN = 'Basenames'
        else:   
            tagN = 'Providename'
        for h in self.ts.dbMatch(tagN, name):
            # need to do version compare here
            # then populate providesCache 
            self.providesCache[dep] = 1
            return 1
        else:   
            return None

        
senses = [rpm.RPMSENSE_CONFIG, rpm.RPMSENSE_EQUAL, rpm.RPMSENSE_FIND_PROVIDES, rpm.RPMSENSE_FIND_REQUIRES, rpm.RPMSENSE_GREATER, rpm.RPMSENSE_INTERP, rpm.RPMSENSE_KEYRING, rpm.RPMSENSE_LESS, rpm.RPMSENSE_PATCHES, rpm.RPMSENSE_PREREQ, rpm.RPMSENSE_RPMLIB, rpm.RPMSENSE_SCRIPT_BUILD, rpm.RPMSENSE_SCRIPT_CLEAN, rpm.RPMSENSE_SCRIPT_INSTALL, rpm.RPMSENSE_SCRIPT_POST, rpm.RPMSENSE_SCRIPT_POSTUN, rpm.RPMSENSE_SCRIPT_PRE, rpm.RPMSENSE_SCRIPT_PREP, rpm.RPMSENSE_SCRIPT_PREUN, rpm.RPMSENSE_SCRIPT_VERIFY, rpm.RPMSENSE_SERIAL, rpm.RPMSENSE_TRIGGERIN, rpm.RPMSENSE_TRIGGERPOSTUN, rpm.RPMSENSE_TRIGGERPREIN, rpm.RPMSENSE_TRIGGERUN]

sensehash = {}
for i in dir(rpm):
    if i[:8] == "RPMSENSE":
        sensehash[eval("rpm.%s" % i)] = "%s" % i

def findsense(sense):
    h = []
    for s in senses:
        if s & sense:
            h.append(sensehash[s])
    return h

def main():
    globstring = sys.argv[1]

#    rpmobj = RpmClass(filename=filename)

#    print rpmobj.getRequires()
#    print rpmobj.getProvides()
#    print rpmobj.getObsoletes()
#    print rpmobj.getConflicts()
#    print rpm.getFiles()

    blip = VTransaction()
    ts = rpm.TransactionSet()
    import glob

    
    filenames = glob.glob(globstring)
    for filename in filenames:
        print "loading: %s" % filename
        fd = os.open(filename, os.O_RDONLY)
        hdr = ts.hdrFromFdno(fd)
        blip.addInstall(hdr, hdr, "i")
        os.close(fd)
        del hdr
        
    print "Requires"
#    print blip.pkgToRequires
#    print blip.requireToPkgs
    print

    print "Provides"
#    print blip.pkgToProvides
#    import pprint
#    pprint.pprint(blip.provideNameToProvideTuple)
#    print blip.provideToPkgs
    print

    print "Obsoletes"
#    print blip.pkgToObsoletes
#    print blip.obsoleteToPkgs
    print

    print "Conflicts"
#    print blip.pkgToConflicts
#    print blip.conflictToPkgs
    print 

#    print blip.findUnsolvedRequires()

    print
#    print blip.providesCache
    foo = blip.genDepsList()
    print foo

    for bar in foo:
        dep = bar[1]
        sense = dep[1]
        print " %s  %s" % (dep, findsense(sense))
        
#   print blip.provideToPkgs.keys()
#    print blip.provideNameToProvideTuple
    
if __name__ == "__main__":
    main()
